"""Unit tests for the hardware_component module."""

import unittest
import time
import threading

from turboctl.telegram.codes import ControlBits, StatusBits
from turboctl.telegram.datatypes import Uint
from turboctl.telegram.parser import PARAMETERS
from turboctl.telegram.telegram import TelegramBuilder, TelegramReader
from turboctl.virtualpump.hardware_component import HardwareComponent
from turboctl.virtualpump.parameter_component import ExtendedParameters

# The attributes of HardwareComponent or the on and off methods aren't
# explicitly tested here, but they are used internally by the class,
# so these tests should fail if there was something wrong with them.
# Likewise, the other classes in the hardware_component module aren't tested,
# because they are only used by HardwareComponent.


class TestStop(unittest.TestCase):
    """Test the stop() method."""
    
    def test_stop(self):
        # Test classes are executed in alphabetical order, so this one is
        # executed last. Because of this, we have to wait for the threads
        # created by the other classes to stop.
        # Even if there was no possibility of dangling parallel threads left by
        # this module, this module might be run just after another test module,
        # which could have created its own parallel threads.
        # 0.2 is twice the default timestep for the parallel thread of
        # HardwareComponent.
        time.sleep(0.2)
        
        # Count active threads.
        original_count = threading.active_count()
        
        # Creating a HardwareComponent should add one thread.
        hwc = HardwareComponent(ExtendedParameters(PARAMETERS),
                                threading.Lock())
        self.assertEqual(threading.active_count(), original_count + 1)
        
        # Calling stop() should reduce the number of threads back to its
        # original value.
        hwc.stop()
        time.sleep(2 * hwc.step)
        self.assertEqual(threading.active_count(), original_count)


class Base(unittest.TestCase):

    def setUp(self):
        self.parameters = ExtendedParameters(PARAMETERS)
        # The lock would normally be used by VirtualPump to prevent the
        # frequency-updating loop from running while the reply is being
        # composed.
        # Since HardwareComponent is tested here alone without a VirtualPump,
        # the lock object doesn't really do anything.
        # Because the lock doesn't work correctly, there is a small chance of a
        # test failing due to a race condition.
        # E.g. when the pump is turned on, the reply should report the
        # frequency as 0, but the parallel loop might actually have time to
        # update the frequency to a nonzero value after the pump has turned on
        # but before the frequency has been read and written to the reply.
        # However, in practice this chance is very small, since the default
        # timestep of the frequency loop is 100 ms, which is very large
        # compared to the time required to execute a few lines of code.
        # In the unlikely event that the tests fail due to this, they may be
        # simply run again.
        self.hwc = HardwareComponent(self.parameters, threading.Lock())
        
    def tearDown(self):
        self.hwc.stop()
        
    def send(self, flag_bits):
        builder = TelegramBuilder().set_flag_bits(flag_bits)

        query = TelegramReader(builder.build('query'), 'query')
        # The reply initially contains the same data as the query, so we can
        # use the builder used to build the query as a basis for the reply.
        self.hwc.handle_hardware(query, builder)
        reply = TelegramReader(builder.build('reply'), 'reply')

        return reply
    
    def pump_off_test(self, reply):
        """Test the status of the pump in the off state."""
        self.assertEqual(reply.flag_bits, [StatusBits.READY,
                                            StatusBits.PARAM_CHANNEL])

        self.assertEqual(reply.frequency, 0)
        self.assertEqual(reply.temperature, 0)
        self.assertEqual(reply.current, 0)
        self.assertEqual(reply.voltage, 0)
        
        self.assertEqual(self.parameters[3].value[0].value, 0)
        self.assertEqual(self.parameters[11].value[0].value, 0)
        self.assertEqual(self.parameters[5].value[0].value, 0)
        self.assertEqual(self.parameters[4].value[0].value, 0)
        
    def on_command_test(self, reply):
        """Test the reply to an on commend."""
        self.assertEqual(set(reply.flag_bits), set([StatusBits.OPERATION,
                                                    StatusBits.ACCELERATION,
                                                    StatusBits.PARAM_CHANNEL,
                                                    StatusBits.PROCESS_CHANNEL
                                                    ]))

        self.assertEqual(reply.frequency, 0)
        self.assertEqual(reply.temperature, HardwareComponent.TEMPERATURE)
        self.assertEqual(reply.current, HardwareComponent.CURRENT)
        self.assertEqual(reply.voltage, HardwareComponent.VOLTAGE)
        
        self.assertEqual(self.parameters[3].value[0].value, 0)
        self.assertEqual(
            self.parameters[11].value[0].value, HardwareComponent.TEMPERATURE)
        self.assertEqual(
            self.parameters[5].value[0].value, HardwareComponent.CURRENT)
        self.assertEqual(
            self.parameters[4].value[0].value, HardwareComponent.VOLTAGE)
        
    def after_on_command_test(self, reply):
        """Test the status of the pump slightly after an on command."""
        self.assertEqual(set(reply.flag_bits), set([StatusBits.OPERATION,
                                                    StatusBits.ACCELERATION,
                                                    StatusBits.PARAM_CHANNEL,
                                                    StatusBits.PROCESS_CHANNEL,
                                                    StatusBits.TURNING
                                                    ]))

        self.assertGreater(reply.frequency, 0)
        self.assertEqual(reply.temperature, HardwareComponent.TEMPERATURE)
        self.assertEqual(reply.current, HardwareComponent.CURRENT)
        self.assertEqual(reply.voltage, HardwareComponent.VOLTAGE)
        
        self.assertGreater(self.parameters[3].value[0].value, 0)
        self.assertEqual(
            self.parameters[11].value[0].value, HardwareComponent.TEMPERATURE)
        self.assertEqual(
            self.parameters[5].value[0].value, HardwareComponent.CURRENT)
        self.assertEqual(
            self.parameters[4].value[0].value, HardwareComponent.VOLTAGE)
        
    def long_after_on_command_test(self, reply):
        """Test the status of the pump after the frequency setpoint has been
        reached."""
        self.assertEqual(set(reply.flag_bits), set([StatusBits.OPERATION,
                                                    StatusBits.PARAM_CHANNEL,
                                                    StatusBits.PROCESS_CHANNEL,
                                                    StatusBits.TURNING
                                                    ]))

        setpoint = self.parameters[24].value[0].value
        
        self.assertEqual(reply.frequency, setpoint)
        self.assertEqual(reply.temperature, HardwareComponent.TEMPERATURE)
        self.assertEqual(reply.current, HardwareComponent.CURRENT)
        self.assertEqual(reply.voltage, HardwareComponent.VOLTAGE)
        
        self.assertEqual(self.parameters[3].value[0].value, setpoint)
        self.assertEqual(
            self.parameters[11].value[0].value, HardwareComponent.TEMPERATURE)
        self.assertEqual(
            self.parameters[5].value[0].value, HardwareComponent.CURRENT)
        self.assertEqual(
            self.parameters[4].value[0].value, HardwareComponent.VOLTAGE)
        
    def off_command_test(self, reply):
        """Test the reply to an off commend."""
        self.assertEqual(set(reply.flag_bits), set([StatusBits.READY,
                                                    StatusBits.TURNING,
                                                    StatusBits.DECELERATION,
                                                    StatusBits.PARAM_CHANNEL,
                                                    ]))

        setpoint = self.parameters[24].value[0].value
        
        self.assertEqual(reply.frequency, setpoint)
        self.assertEqual(reply.temperature, 0)
        self.assertEqual(reply.current, 0)
        self.assertEqual(reply.voltage, 0)
        
        self.assertEqual(self.parameters[3].value[0].value, setpoint)
        self.assertEqual(self.parameters[11].value[0].value, 0)
        self.assertEqual(self.parameters[5].value[0].value, 0)
        self.assertEqual(self.parameters[4].value[0].value, 0)
        
    def after_off_command_test(self, reply):
        """Test the status of the pump slightly after an off command."""
        self.assertEqual(set(reply.flag_bits), set([StatusBits.READY,
                                                    StatusBits.TURNING,
                                                    StatusBits.DECELERATION,
                                                    StatusBits.PARAM_CHANNEL,
                                                    ]))

        setpoint = self.parameters[24].value[0].value
        
        self.assertLess(reply.frequency, setpoint)
        self.assertEqual(reply.temperature, 0)
        self.assertEqual(reply.current, 0)
        self.assertEqual(reply.voltage, 0)
        
        self.assertLess(self.parameters[3].value[0].value, setpoint)
        self.assertEqual(self.parameters[11].value[0].value, 0)
        self.assertEqual(self.parameters[5].value[0].value, 0)
        self.assertEqual(self.parameters[4].value[0].value, 0)

        
class TestHardwareComponent(Base):
    
    def test_cycle(self):
        """Test a complete on-off cycle."""
        
        # Ask pump status in the off state.
        reply = self.send(flag_bits=[])
        self.pump_off_test(reply)
        
        # Send the on command and check the response.
        reply = self.send(flag_bits=[ControlBits.COMMAND, ControlBits.ON])
        self.on_command_test(reply)
        
        # Wait for the pump to start, and ask its status.
        # The wait time is twice the timestep of the pump so that it certainly
        # has enough time to update the frequency.
        time.sleep(2 * self.hwc.step)
        reply = self.send(flag_bits=[ControlBits.COMMAND, ControlBits.ON])
        self.after_on_command_test(reply)
    
        # Set the acceleration rate to infinite, wait for the frequency to
        # reach the setpoint, and ask pump status.
        # Afterwards the acceleration rate is reset back to its old value.
        old_value = self.hwc.abs_acceleration
        self.hwc.abs_acceleration = float('inf')
        time.sleep(2 * self.hwc.step)
        reply = self.send(flag_bits=[ControlBits.COMMAND, ControlBits.ON])
        self.long_after_on_command_test(reply)
        self.hwc.abs_acceleration = old_value
    
        # Send the off command and check the response.
        reply = self.send(flag_bits=[])
        self.off_command_test(reply)
    
        # Wait for the pump to turn off, and ask its status.
        time.sleep(2 * self.hwc.step)
        reply = self.send(flag_bits=[])
        self.after_off_command_test(reply)
        
        # Set the acceleration to infinite again, wait for the requency to
        # drop to zero, and ask pump status.
        self.hwc.abs_acceleration = float('inf')
        time.sleep(2 * self.hwc.step)
        reply = self.send(flag_bits=[])
        self.pump_off_test(reply)
        
    def test_change_setpoint(self):
        self.hwc.abs_acceleration = float('inf')

        # Changing the setpoint while the pump is off shouldn't affect the
        # frequency immediately.
        self.parameters[24].value[0] = Uint(1234, 16)
        time.sleep(2 * self.hwc.step)
        reply = self.send(flag_bits=[])
        self.assertEqual(reply.frequency, 0)
        
        # However, after sending the on command the frequency becomes equal to
        # the checkpoint.
        # Send the on command.
        self.send(flag_bits=[ControlBits.COMMAND, ControlBits.ON])
        time.sleep(2 * self.hwc.step)
        # Ask for pump status.
        reply = self.send(flag_bits=[ControlBits.COMMAND, ControlBits.ON])
        self.assertEqual(reply.frequency, 1234)
        
        # Change the setpoint to a bigger value.
        self.parameters[24].value[0] = Uint(4567, 16)
        time.sleep(2 * self.hwc.step)
        reply = self.send(flag_bits=[ControlBits.COMMAND, ControlBits.ON])
        self.assertEqual(reply.frequency, 4567)
        
        # Change the setpoint to a smaller value.
        self.parameters[24].value[0] = Uint(89, 16)
        time.sleep(2 * self.hwc.step)
        reply = self.send(flag_bits=[ControlBits.COMMAND, ControlBits.ON])
        self.assertEqual(reply.frequency, 89)
        
    def test_on_control_bit_alone(self):
        """Make sure the 'ON' control bit doesn't do anything without the
        'COMMAND' control bit.
        """
        # Sending the 'ON' bit in the off state doesn't turn the pump on.
        reply = self.send(flag_bits=[ControlBits.ON])
        self.assertEqual(set(reply.flag_bits), set([StatusBits.READY,
                                                    StatusBits.PARAM_CHANNEL]))
        
        # Turn the pump on.
        self.send(flag_bits=[ControlBits.COMMAND, ControlBits.ON])
        self.assertTrue(self.hwc.is_on)
        
        # Sending the 'ON' bit in the on state turns the pump off just like an
        # empty telegram does.
        reply = self.send(flag_bits=[ControlBits.ON])
        self.assertFalse(self.hwc.is_on)
        self.assertEqual(set(reply.flag_bits), set([StatusBits.READY,
                                                    StatusBits.PARAM_CHANNEL]))
        
    def test_command_control_bit_alone(self):
        """Make sure the 'COMMAND' control bit doesn't do anything (except add
        the 'PROCESS_CHANNEL' status bit) without the 'ON' control bit.
        """
        # Sending the 'COMMAND' bit in the off state doesn't turn the pump on.
        reply = self.send(flag_bits=[ControlBits.COMMAND])
        self.assertEqual(set(reply.flag_bits), set([StatusBits.READY,
                                                    StatusBits.PARAM_CHANNEL,
                                                    StatusBits.PROCESS_CHANNEL
                                                    ]))
        
        # Turn the pump on.
        self.send(flag_bits=[ControlBits.COMMAND, ControlBits.ON])
        
        # Sending the 'COMMAND' bit in the on state turns the pump off just
        # like an empty telegram does.
        reply = self.send(flag_bits=[ControlBits.COMMAND])
        self.assertFalse(self.hwc.is_on)
        self.assertEqual(set(reply.flag_bits), set([StatusBits.READY,
                                                    StatusBits.PARAM_CHANNEL,
                                                    StatusBits.PROCESS_CHANNEL
                                                    ]))
    
        
if __name__ == '__main__':
    unittest.main()



"""Unit tests for the parser module."""

import unittest

from turboctl.telegram.datatypes import Uint, Sint, Float
from turboctl.telegram.parser import (
    Parameter, ErrorOrWarning, PARAMETERS, ERRORS, WARNINGS, _parse as parse)

# Attributes of the Parameter class:
# number, name, indices, min, max, default, unit, writable, datatype, bits, 
# description

# Attributes of the ErrorOrWarning class: number, name, possible_cause, remedy

def dummy_parameter(number=1, 
                    name='Test parameter',  
                    indices=range(0), 
                    min_value=Uint(0, 16), 
                    max_value=Uint(65535, 16),
                    default=Uint(0, 16), 
                    unit='', 
                    writable=True, 
                    datatype=Uint, 
                    bits=16, 
                    description='Test description.'):
    """Create a Parameter object where all fields take default values unless 
    otherwise specified in the arguments.
    """
    return Parameter(number=number,
                     name=name,
                     indices=indices, 
                     min_value=min_value, 
                     max_value=max_value, 
                     default=default, 
                     unit=unit, 
                     writable=writable, 
                     datatype=datatype,
                     bits=bits,
                     description=description)
    
def dummy_parameter_from_line(number='1', 
                              name='"Test parameter"',
                              min_value='0', 
                              # This max_value is small enough for Sints.
                              max_value='32767',
                              default='0', 
                              unit='""', 
                              rw='r/w', 
                              format_='u16',
                              description='"Test description."'):
    """Create a parameter object by parsing a line of text."""
    
    string = ' '.join([number, name, min_value, max_value, default, unit, rw, 
                       format_, description])
    return parse(string, 'parameter')
    

class TestParsing(unittest.TestCase):
    
    def test_general(self):
        line = '1 "Test parameter" 0 65535 0 "" r/w u16 "Test description."'
        parameter = parse(line, 'parameter')
        self.assertEqual(parameter, dummy_parameter())
        
    def test_too_many_fields(self):
        line = '1 "Test parameter" 0 65535 0 "" "" r/w u16 "Test description."'
        with self.assertRaises(ValueError):
            parse(line, 'parameter')
            
    def test_too_few_fields(self):
        line = '1 "Test parameter" 0 65535 0 r/w u16 "Test description."'
        with self.assertRaises(ValueError):
            parse(line, 'parameter')

    def test_name_without_quotes(self):
        parameter = dummy_parameter_from_line(name='name')
        self.assertEqual(parameter.name, 'name')

    def test_name_with_line_break(self):
        """Make sure a literal \n is turned into the newline 
        character.
        """
        parameter = dummy_parameter_from_line(name=r'"Lorem\nipsum"')
        self.assertEqual(parameter.name, 'Lorem\nipsum')
        self.assertNotEqual(parameter.name, r'Lorem\nipsum')

    def test_unindexed(self):
        parameter = dummy_parameter_from_line(number='1')
        self.assertEqual(parameter.indices, range(0))
        
    def test_indexed(self):
        parameter = dummy_parameter_from_line(number='1[1:5]')
        self.assertEqual(parameter.indices, range(1,6))
  
    def test_min_max_int(self):
        parameter = dummy_parameter_from_line(
            min_value='-10', max_value='100', format_='s16')
        self.assertEqual(parameter.min_value, Sint(-10, 16))
        self.assertEqual(parameter.max_value, Sint(100, 16))
        
    def test_min_max_float(self):
        parameter = dummy_parameter_from_line(
            min_value='-1.23e-4', max_value='4.56e7', format_='real32')
        self.assertEqual(parameter.min_value, Float(-1.23e-4))
        self.assertEqual(parameter.max_value, Float(4.56e7))
        
    def test_min_max_reference(self):
        parameter = dummy_parameter_from_line(
            min_value='P1', max_value='P2')
        self.assertEqual(parameter.min_value, 'P1')
        self.assertEqual(parameter.max_value, 'P2')
        
        with self.assertRaises(ValueError):
            dummy_parameter_from_line(min_value='PX')
            
    def test_default_int(self):
        parameter = dummy_parameter_from_line(default='10')
        self.assertEqual(parameter.default, Uint(10, 16))
        
        parameter = dummy_parameter_from_line(
            default='-10', format_='s16')
        self.assertEqual(parameter.default, Sint(-10, 16))
        
    def test_default_float(self):
        parameter = dummy_parameter_from_line(
            default='1.2e3', format_='real32')
        self.assertEqual(parameter.default, Float(1.2e3))

    def test_indexed_default_with_a_single_value(self):
        parameter = dummy_parameter_from_line(number='1[1:3]', default='3')
        self.assertEqual(parameter.default, Uint(3, 16))
        
    def test_indexed_default_with_multiple_values(self):
        parameter = dummy_parameter_from_line(
            number='1[1:3]', default='[3,2,1]')
        
        self.assertEqual(
            parameter.default, [Uint(3, 16), Uint(2, 16), Uint(1, 16)])
        
    def test_unit(self):
        parameter = dummy_parameter_from_line(unit='"0.1 °C"')
        self.assertEqual(parameter.unit, '0.1 °C')
        
    def test_writable(self):
        parameter = dummy_parameter_from_line(rw='r/w')
        self.assertTrue(parameter.writable)
        
    def test_not_writable(self):
        parameter = dummy_parameter_from_line(rw='""')
        self.assertFalse(parameter.writable)
        
        parameter = dummy_parameter_from_line(rw='r')
        self.assertFalse(parameter.writable)
        
    def test_invalid_writability_fails(self):
        
        with self.assertRaises(ValueError):
            dummy_parameter_from_line(rw='w')
            
        with self.assertRaises(ValueError):
            dummy_parameter_from_line(rw='xyz')
            
    def test_uint(self):
        formats = ['u16', 'u32']
        
        for f in formats:
            with self.subTest(i=f):
                parameter = dummy_parameter_from_line(format_=f)
                self.assertEqual(parameter.datatype, Uint)
    
    def test_sint(self):
        formats = ['s16', 's32']
        
        for f in formats:
            with self.subTest(i=f):
                parameter = dummy_parameter_from_line(format_=f)
                self.assertEqual(parameter.datatype, Sint)
        
    def test_float(self):
        f = 'real32'
        parameter = dummy_parameter_from_line(format_=f)
        self.assertEqual(parameter.datatype, Float)
            
    def test_16_bits(self):
        formats = ['u16', 's16']
        
        for f in formats:
            with self.subTest(i=f):
                parameter = dummy_parameter_from_line(format_=f)
                self.assertEqual(parameter.bits, 16)
        
    def test_32_bits(self):
        formats = ['u32', 's32', 'real32']
        
        for f in formats:
            with self.subTest(i=f):
                parameter = dummy_parameter_from_line(format_=f)
                self.assertEqual(parameter.bits, 32)
                
    def test_invalid_bits_fails(self):
        formats = ['16', 'x16', 'u']
        
        for f in formats:
            with self.subTest(i=f):
                with self.assertRaises(ValueError):
                    dummy_parameter_from_line(format_=f)
                
    def test_description_without_quotes(self):
        parameter = dummy_parameter_from_line(description='test')
        self.assertEqual(parameter.description, 'test')

    def test_description_with_line_break(self):
        """Make sure a literal \n is turned into the newline 
        character.
        """
        parameter = dummy_parameter_from_line(description=r'"Lorem\nipsum"')
        self.assertEqual(parameter.description, 'Lorem\nipsum')
        self.assertNotEqual(parameter.description, r'Lorem\nipsum')   
        
        
class TestActualParameters(unittest.TestCase):
    """Make sure the parameters defined in parameters.txt aren't blatantly 
    incorrect.
    """
    
    def test_keys_equal_numbers(self):
        for num, parameter in PARAMETERS.items():
            self.assertEqual(PARAMETERS[num].number, num)
            
    def test_all_parameters_are_defined(self):
        numbers = {1,2,3,4,5,6,7,8,11,16,17,18,19,20,21,23,24,25,26,27,28,29,
                   30,31,32,36,37,38,40,41,43,119,122,125,126,128,131,132,133,
                   134,140,150,171,174,176,179,180,182,183,184,185,227,247,
                   248,249,312,313,314,315,316,349,350,355,394,395,396,601,
                   602,604,606,609,610,611,615,616,617,618,619,620,623,624,
                   625,634,636,643,644,647,648,649,652,670,671,672,673,678,
                   679,682,686,690,918,923,924,1025,1035,1100,1101,1102}
        self.assertEqual(PARAMETERS.keys(), numbers)
        
    # These two tests together also assert that all parameter numbers 
    # are ints (since all keys are ints and parameter numbers are equal 
    # to keys).
    
    def test_all_names_are_strings(self):
        for parameter in PARAMETERS.values():
            self.assertIsInstance(parameter.number, int)
            
    def test_all_indices_are_ranges(self):
        for parameter in PARAMETERS.values():
            self.assertIsInstance(parameter.indices, range)
            
    def test_all_mins_are_data_or_references(self):
        
        for parameter in PARAMETERS.values():        
            value = parameter.min_value
            
            # Convert a string value 'P<number>' to an Uint:
            try:
                if value[0] == 'P':
                    value = Uint(int(value[1:]))
            except (TypeError, ValueError):
                pass
            
            self.assertIsInstance(value, (Uint, Sint, Float))
            
    def test_all_maxes_are_data_or_references(self):
        
        for parameter in PARAMETERS.values():
            value = parameter.max_value
            
            # Convert a string value 'P<number>' to an Uint:
            try:
                if value[0] == 'P':
                    value = Uint(int(value[1:]))
            except (TypeError, ValueError):
                pass
            
            self.assertIsInstance(value, (Uint, Sint, Float))  

    def test_all_defaults_are_data_or_list_of_data(self):
        
        for parameter in PARAMETERS.values():
            
            value = parameter.default
            value_is_data = type(value) in (Uint, Sint, Float)
            value_is_list =  type(value) == list
            try:
                values_are_data = all([type(i) in (Uint, Sint, Float) 
                                       for i in value])
            except TypeError:
                # Data instances aren't iterable and will raise this
                # error.
                pass
                            
            self.assertTrue(
                    value_is_data or (value_is_list and values_are_data)) 
            
    def test_all_units_are_strings(self):
        for parameter in PARAMETERS.values():
            self.assertIsInstance(parameter.unit, str) 

    def test_all_writable_values_are_booleans(self):
        for parameter in PARAMETERS.values():
            self.assertIsInstance(parameter.writable, bool)

    def test_all_datatypes_are_uint_sint_or_float(self):
        for parameter in PARAMETERS.values():
            self.assertTrue(
                    parameter.datatype in (Uint, Sint, Float))

    def test_all_bits_are_16_or_32(self):
        for parameter in PARAMETERS.values():
            self.assertTrue(parameter.bits in (16, 32))

    def test_all_descriptions_are_strings(self):
        for parameter in PARAMETERS.values():
            self.assertIsInstance(parameter.description, str)
            
            
class TestErrorsAndWarnings(unittest.TestCase):
    
    def test_parse_warning(self):
        line = '0 "Test warning" "Test cause" "Test remedy"'
        warning = parse(line, 'warning')
        self.assertEqual(warning, ErrorOrWarning(
            0, "Test warning", "Test cause", "Test remedy"))
        
    def test_parse_error(self):
        line = '0 "Test error" "Test cause" "Test remedy"'
        error = parse(line, 'error')
        self.assertEqual(error, ErrorOrWarning(
            0, "Test error", "Test cause", "Test remedy"))
        
    def test_all_warnings_defined(self):
        warning_numbers = {0,1,2,3,6,7,11,12,13,14}
        self.assertEqual(set(WARNINGS.keys()), warning_numbers)
        
    def test_all_errors_defined(self):
        error_numbers = {
            1,2,3,4,5,6,7,8,61,82,83,84,*range(85,97),97,101,103,106,111,116,
            117,126,128,143,144,225,*range(226,237),237,238,240,252,600,601,
            602,608,609,603,610,611,612
        }
        self.assertEqual(set(ERRORS.keys()), error_numbers)
        
    def test_all_warning_fields_strings(self):
        for w in WARNINGS.values():
            self.assertIsInstance(w.name, str)
            self.assertIsInstance(w.possible_cause, str)
            self.assertIsInstance(w.remedy, str)
            
    def test_all_error_fields_strings(self):
        for w in ERRORS.values():
            self.assertIsInstance(w.name, str)
            self.assertIsInstance(w.possible_cause, str)
            self.assertIsInstance(w.remedy, str)

          
if __name__ == '__main__':
    unittest.main()


"""Unit tests for the codes module. 

Since the codes module contains mostly data, these tests only make 
sure that the superclasses ValueAndDescription and IntAndDescription 
defined in that module work as intended.
"""
import unittest

from turboctl.telegram.codes import (ParameterAccess, ParameterResponse, 
                                     get_parameter_code, get_parameter_mode,
                                     ParameterException, WrongNumError,
                                     ParameterError, ControlBits ,StatusBits)


class TestParameterCode(unittest.TestCase):
    
    def test_syntax(self):
        """Make sure enum members can be accessed using any valid 
        syntax.
        """
        self.assertEqual(ParameterAccess['NONE'], ParameterAccess.NONE)
        self.assertEqual(ParameterAccess('0000'), ParameterAccess.NONE)
        self.assertEqual(ParameterAccess(ParameterAccess.NONE), 
                         ParameterAccess.NONE)
        
    def test_fields(self):
        """Make sure both attributes can be accessed."""
        self.assertEqual(ParameterAccess.NONE.value, '0000')
        self.assertEqual(ParameterAccess.NONE.mode, 'none')
        self.assertEqual(ParameterAccess.NONE.indexed, ...)
        self.assertEqual(ParameterAccess.NONE.bits, ...)
        self.assertEqual(ParameterAccess.NONE.description, 'No access')


class TestGetParameterCode(unittest.TestCase):
    
    def test_examples(self):
        code = get_parameter_code('query', 'none', False, 16)
        self.assertEqual(code, ParameterAccess.NONE)
        
        code = get_parameter_code('reply', 'none', False, 16)
        self.assertEqual(code, ParameterResponse.NONE)

        code = get_parameter_code('query', 'write', False, 16)
        self.assertEqual(code, ParameterAccess.W16)
        
        code = get_parameter_code('reply', 'response', True, 32)
        self.assertEqual(code, ParameterResponse.S32F)

        code = get_parameter_code('reply', 'error', False, 16)
        self.assertEqual(code, ParameterResponse.ERROR)

    def test_invalid_mode(self):
        with self.assertRaises(ValueError):
            get_parameter_code('qery', 'none', False, 16)
        
    def test_no_matches(self):
        with self.assertRaises(ValueError):
            get_parameter_code('query', 'response', False, 16)
            
            
class TestGetParameterMode(unittest.TestCase):
    
    def test_examples(self):
        mode = get_parameter_mode('query', '0001')
        self.assertEqual(mode, 'read')
        
        mode = get_parameter_mode('reply', '0001')
        self.assertEqual(mode, 'response')

    def test_invalid_mode(self):
        with self.assertRaises(ValueError):
            get_parameter_mode('qery', '0000')
        
    def test_no_matches(self):
        with self.assertRaises(ValueError):
            get_parameter_mode('query', '1111')

        
class TestCustomIntEnums(unittest.TestCase):
    """Make sure inheriting CustomInt makes enum members behave like ints."""
        
    def test_set(self):
        """Make sure sets of enum members behave as they should."""
        
        # Names of enum members in the correct order.
        bitnames = [
            'ON', 'UNUSED1', 'UNUSED2', 'UNUSED3', 'UNUSED4', 'X201', 
            'SETPOINT', 'RESET_ERROR', 'STANDBY', 'UNUSED9', 'COMMAND', 
            'X1_ERROR', 'X1_WARNING', 'X1_NORMAL', 'X202', 'X203'
        ]
        # Numbers of enum members in order.
        bits_in_order = [ControlBits[name] for name in bitnames]

        # Print long error messages so that the order of all bits can be seen.  
        self.maxDiff = None
        
        # Unsort the bits by reversing their order.
        bits_reversed = bits_in_order[::-1]
        # list() automatically sorts a set of ints.
        bits_should_be_in_order = list(set(bits_reversed))
        
        # Make sure the bits are in the correct order and all of them 
        # are present (i.e. hashing works)
        self.assertEqual(bits_should_be_in_order, bits_in_order)
        self.assertEqual(bits_should_be_in_order, list(range(16)))
        
    def test_equal(self):
        """Make sure enum members are equal to ints."""
        self.assertEqual(ControlBits.ON, 0)
        self.assertEqual(ControlBits.COMMAND, 10)
        
    def test_add(self):
        """Make sure enum members are compatible with arithmetical 
        operations.
        """
        self.assertEqual(ControlBits.COMMAND + ControlBits.ON, 10)


class TestCustomIntEnumFields(unittest.TestCase):
    """Make sure ParameterError, ControlBits an FlagBits members function and
    have the fields they should have.
    """
    
    def test_parameter_error(self):
        member = ParameterError.WRONG_NUM
        self.assertEqual(member, 0)
        self.assertEqual(member.value, 0)
        self.assertEqual(member.description, 'Invalid parameter number')
        with self.assertRaises(WrongNumError):
            raise member.exception('Test')
            
    def test_control_bits(self):
        member = ControlBits.ON
        self.assertEqual(member, 0)
        self.assertEqual(member.value, 0)
        self.assertEqual(member.description, 'Turn or keep the pump on')

    def test_status_bits(self):
        member = StatusBits.READY
        self.assertEqual(member, 0)
        self.assertEqual(member.value, 0)
        self.assertEqual(member.description, 'Ready for operation')
        
        
class TestParameterException(unittest.TestCase):
    """Make sure ParameterException.member attributes work properly."""
    
    def test_exception_member(self):
        for member in ParameterError:
            self.assertEqual(member.exception.member, member)
            
    def test_superclass_member(self):
        """The ParameterException superclass shouldn't have the member
        attribute.
        """
        with self.assertRaises(AttributeError):
            ParameterException.member

        
if __name__ == '__main__':
    unittest.main()